// LED pendant controller
//
// Designed for an ATTINY45/85 with the following connections:
//
//                                        +3v
//                              ______     T
//                        /---=|1 \/ 8|=---+
//                        |   =|2    7|=   V <--red LED
//         microswitch -|     =|3    6|=   |
//                        +---=|4____5|=---/
//                        |
//                       GND
//
// +3v to VCC (physical pin 8)
// GND to GND (physical pin 4)
// Microswitch between GND and RESET (physical pin 1)
// LED (red works best) between VCC and PIN 0 (physical pin 5)

#include <avr/sleep.h>
#include <EEPROM.h>
#include "sequences.h"



// The brightest value for the LED.  Lower values result in better battery life while higher values result in brighter light
const byte MAX_LED_VALUE = 255;

// The number of animation frames per second
const byte FRAMES_PER_SECOND = 20;

// A delay used to "debounce" the reset switch
const unsigned long STARTUP_DEBOUNCE_DELAY_MS = 50;

// A double-click of the RESET switch causes the microcontroller to power off immediately
const unsigned long DOUBLE_CLICK_DELAY_MS = 1000;

// The pendant will automatically go to sleep after this many minutes
const unsigned long SLEEP_TIMEOUT_MIN = 10;

// Pins
const byte LED_PIN = 0; // Physical pin 5

// EEPROM addresses
enum class EEPROM_ADDRESSES
{
  INITIALIZED_FLAG,
  DOUBLE_CLICK_FLAG,
  SLEEP_FLAG,
  CURRENT_SEQUENCE_NUMBER,
};
const byte MAGIC_NUMBER = 0x42;
const byte FLAG_SET_VALUE = 0x00;
const byte FLAG_UNSET_VALUE = 0xFF;

// Calculated constants
const unsigned long FRAME_DELAY_MS = 1000 / FRAMES_PER_SECOND;
const unsigned long DOUBLE_CLICK_DELAY_FRAMES = DOUBLE_CLICK_DELAY_MS / FRAME_DELAY_MS;
const unsigned long SLEEP_TIMEOUT_FRAMES = (SLEEP_TIMEOUT_MIN * 60 * 1000) / FRAME_DELAY_MS;

// Global variables
unsigned long g_totalFrames = 0;
byte g_sequenceNumber = 0;
unsigned int g_sequenceStep = 0;



// Put the processor into sleep mode
void GoToSleep()
{
  // Record that the pendant was put to sleep
  EEPROM.update (static_cast<byte>(EEPROM_ADDRESSES::SLEEP_FLAG), FLAG_SET_VALUE);
  
  // SLEEP_MODE_PWR_DOWN provides the most energy savings
  set_sleep_mode(SLEEP_MODE_PWR_DOWN);
  sleep_enable();
  sleep_mode();

  // Nothing further needs to be done, since the processor can only be woken by a LOW
  // on RESET, which will start execution over from the beginning
} // GoToSleep()



// Set the light value of the LED
// inputValue should be 0-255 and is rescaled to a range from 0 to MAX_LED_VALUE.
// Also, because the LED is connected to VCC and is grounded by the PWM signal, the output
// value needs to be inverted.
void SetLedValue(byte inputValue)
{
  // Rescale the input value
  byte scaledValue = static_cast<byte>((inputValue / 255.0f) * MAX_LED_VALUE);

  // Invert the value
  byte invertedValue = 255 - scaledValue;
  
  // Set the LED value
  analogWrite(LED_PIN, invertedValue);
} // SetLedValue()



// Process the next frame of the light sequence
void ProcessFrame()
{
  // Determine the next LED value in the sequence
  byte value = pgm_read_byte(SEQUENCE_ADDRESSES[g_sequenceNumber] + g_sequenceStep);

  // Set the LED value
  SetLedValue(value);

  // Determine the next sequence step number
  ++g_sequenceStep;
  if (g_sequenceStep >= SEQUENCE_LENGTHS[g_sequenceNumber])
  {
    // Wrap back to the beginning of the sequence
    g_sequenceStep = 0;
  }
} // ProcessFrame()



// setup is called whenever the RESET switch is pushed
// If the RESET switch is pressed one time, it rotates to the next LED sequence
// If the RESET switch is double-clicked within the double-click timeout period, it puts the pendant to sleep
void setup() 
{
  // Configure the LED pin
  pinMode(LED_PIN, OUTPUT);
  digitalWrite(LED_PIN, HIGH); // Turn the LED OFF
  
  // Delay to debounce the reset pin
  delay(STARTUP_DEBOUNCE_DELAY_MS);

  // If the EEPROM has not been initialized
  if (EEPROM.read(static_cast<byte>(EEPROM_ADDRESSES::INITIALIZED_FLAG)) != MAGIC_NUMBER)
  {
    // Initialize EEPROM with default values
    EEPROM.update(static_cast<byte>(EEPROM_ADDRESSES::INITIALIZED_FLAG), MAGIC_NUMBER);
    EEPROM.update(static_cast<byte>(EEPROM_ADDRESSES::DOUBLE_CLICK_FLAG), FLAG_UNSET_VALUE);
    EEPROM.update(static_cast<byte>(EEPROM_ADDRESSES::SLEEP_FLAG), FLAG_SET_VALUE);
    EEPROM.update(static_cast<byte>(EEPROM_ADDRESSES::CURRENT_SEQUENCE_NUMBER), 0);
  }

  // If the DOUBLE-CLICK flag is set
  // (the DOUBLE-CLICK flag is set during normal startup and unset after the DOUBLE-CLICK timeout
  //  expires.  If the DOUBLE-CLICK flag is set during startup it means that the RESET switch 
  //  was pushed at least twice within the DOUBLE-CLICK timeout, which is treated as a power down command.)
  byte doubleClickFlag = EEPROM.read(static_cast<byte>(EEPROM_ADDRESSES::DOUBLE_CLICK_FLAG));
  if (doubleClickFlag == FLAG_SET_VALUE)
  {
    // Clear the flag
    EEPROM.update(static_cast<byte>(EEPROM_ADDRESSES::DOUBLE_CLICK_FLAG), FLAG_UNSET_VALUE);

    // Put the processor in sleep mode
    GoToSleep();
  } // The DOUBLE-CLICK flag is set

  // Set the double-click flag
  EEPROM.update(static_cast<byte>(EEPROM_ADDRESSES::DOUBLE_CLICK_FLAG), FLAG_SET_VALUE);

  // Determine the current sequence
  g_sequenceNumber = EEPROM.read(static_cast<byte>(EEPROM_ADDRESSES::CURRENT_SEQUENCE_NUMBER));

  // Determine if the pendant is waking up from sleep
  // If waking from sleep, the previous sequence is resumed
  byte sleepFlag = EEPROM.read(static_cast<byte>(EEPROM_ADDRESSES::SLEEP_FLAG));
  EEPROM.update(static_cast<byte>(EEPROM_ADDRESSES::SLEEP_FLAG), FLAG_UNSET_VALUE);

  // If the pendant is not waking up from sleep
  // This indicates the RESET switch was pushed with the intention of switching
  // to the next sequence
  if (sleepFlag == FLAG_UNSET_VALUE)
  {
    ++g_sequenceNumber;
    if (g_sequenceNumber >= NUMBER_OF_SEQUENCES)
    {
      g_sequenceNumber = 0;
    }
    EEPROM.update(static_cast<byte>(EEPROM_ADDRESSES::CURRENT_SEQUENCE_NUMBER), g_sequenceNumber);
  } // The pendant is not waking up from sleep

  // Turn the LED off for now
  SetLedValue(0);
} // setup()



// loop is called repeatedly during normal operation
void loop()
{
  // If the DOUBLE-CLICK delay has been exceeded
  if (g_totalFrames >= DOUBLE_CLICK_DELAY_FRAMES)
  {
    // Clear the DOUBLE-CLICK flag
    EEPROM.update(static_cast<byte>(EEPROM_ADDRESSES::DOUBLE_CLICK_FLAG), FLAG_UNSET_VALUE);
  } // The DOUBLE-CLICK delay has been exceeded

  // If it's time to go to sleep
  if (g_totalFrames >= SLEEP_TIMEOUT_FRAMES)
  {
    GoToSleep();
  } // It's time to go to sleep

  // Process this frame of the light sequence
  ProcessFrame();

  // Increment the total frame count
  ++g_totalFrames;
  
  // Wait until time for the next frame
  delay(FRAME_DELAY_MS);
} // loop()
